import json
import subprocess
import time
import argparse
import os
from termcolor import colored

from secgen.utils import find_available_gpu, all_vuls, vul_to_lang


def fc_infilling_bench_name(lang):
    if "fc_test" in config and config["fc_test"]:
        print("Using fc test set")
        return f"multiple-{lang}_fim_test"
    else:
        return f"multiple-{lang}_fim"


def fc_infilling_command(config, gpu):
    global DEBUG
    command = f"cd ../multipl-e && "
    command += f"CUDA_VISIBLE_DEVICES={gpu} "
    
    if config["no_suffix"]:
        command += f"python fill_in_nosuf.py "
    else:
        command += f"python fill_in.py "
    
    command += f'--model_dir  {config["model_dir"]} '
    command += f'--benchmark {fc_infilling_bench_name(config["lang"])} '
    command += f'--temp {config["temp"]} '
    command += f'--baseline_path {config["baseline_path"]} '
    
    if DEBUG:
        command += f'--n_samples 5 '
    else:
        command += f" >> {config['baseline_path']}/log.txt 2>&1"
    
    return command


def skip_fc_infill_config(config):
    return False
    return os.path.exists(f"{config['baseline_path']}/multiple-{config['lang']}_fim_test.json")

def post_process_fc_infill(fc_infills_test_path, lang):
    data = json.load(open(fc_infills_test_path))
    task_dataset = json.load(open(f"../multipl-e/multiple-{lang}_fim_test.json"))

    print(len(data), len(task_dataset))

    new_data = []
    task_names = []
    current_task_name = None
    current_completion_accumulator = []
    for i in range(len(data)):
        if current_task_name is None:
            current_task_name = task_dataset[i]["name"]
            current_completion_accumulator = data[i]
        elif current_task_name == task_dataset[i]["name"]:
            current_completion_accumulator.extend(data[i])
        else:
            new_data.append(current_completion_accumulator)
            task_names.append(current_task_name)
            current_task_name = task_dataset[i]["name"]
            current_completion_accumulator = data[i]

    new_data.append(current_completion_accumulator)
    task_names.append(current_task_name)

    with open(fc_infills_test_path, "w") as f:
        json.dump(new_data, f, indent=4)

def launch_all_fc_infill(configs, gpus):
    start = time.time()
    print(f"#### Launching fill in on {len(configs)} configurations ####")
    subprocesses = []
    for config in configs:
        if skip_fc_infill_config(config):
            print(
                colored(
                    f"Skipping {config['model_dir']}, {config['lang']}, fc infill already exists",
                    "yellow",
                )
            )
            continue

        gpu = find_available_gpu(gpus)
        print(f"Launching fill in for {config['model_dir']}, {config['lang']} on GPU {gpu}")
        make_dir(config)
        p = subprocess.Popen(fc_infilling_command(config, gpu), shell=True)
        # p.wait()
        subprocesses.append(p)
        time.sleep(60)

    # wait for all launched fc subprocesses to finish
    for p in subprocesses:
        exit_code = p.wait()
        if exit_code != 0:
            print(colored("One of the subprocesses had an error", "red"))

    # post-process results
    for config in configs:
        if config["fc_test"]:
            lang = config["lang"]
            fc_infills_test_path = f"{config['baseline_path']}/multiple-{lang}_fim_test.json"
            post_process_fc_infill(fc_infills_test_path, lang)

    print(f"Fc fill took {time.strftime('%H:%M:%S', time.gmtime(time.time() - start))}")


def fc_measure_task_name(lang):
    return f"multiple-{lang}"


def fc_measure_command(config):
    global DEBUG
    generations_path = config["baseline_path"][3:] + f"/{fc_infilling_bench_name(config['lang'])}.json"
    program_command = f"accelerate launch main.py "
    program_command += f"--tasks {fc_measure_task_name(config['lang'])} "
    program_command += f"--load_generations_path {generations_path} "
    program_command += f"--metric_output_path {generations_path.replace('.json', '.results.json')} "
    program_command += f"--allow_code_execution "
    program_command += f"--n_samples 100 "

    if not DEBUG:
        program_command += f" >> {config['baseline_path'][3:]}/log.txt 2>&1 "

    command = "cd ../../bigcode-evaluation-harness/ && "
    command += f'bash -c "source activate root; conda activate harness; {program_command}"'
    return command


def skip_fc_measure_config(config):
    return False
    return os.path.exists(f"{config['baseline_path']}/multiple-{config['lang']}_fim.results.json")


def launch_all_fc_measure(configs):
    start = time.time()
    print(f"#### Launching measure on {len(configs)} configurations ####")
    for config in configs:
        if skip_fc_measure_config(config):
            print(
                colored(
                    f"Skipping {config['model_dir']}, {config['lang']}, fc result already exists",
                    "yellow",
                )
            )
            continue

        print(f"Launching measure for {config['model_dir']}, {config['lang']}")
        completed_process = subprocess.run(fc_measure_command(config), shell=True)
        if completed_process.returncode != 0:
            print(f"Error in {config['model_dir']}, {config['lang']}")

    print(f"Fc measure took {time.strftime('%H:%M:%S', time.gmtime(time.time() - start))}")


def make_dir(config) -> None:
    directory_path = config["baseline_path"]
    if not os.path.exists(directory_path):
        os.makedirs(directory_path)


temp = 0.4
models = ["bigcode/starcoderbase-3b"]
# ["bigcode/starcoderbase-3b" "codellama/CodeLlama-7b-hf", "gpt-3.5-turbo-instruct-0914", "copilot", "mistralai/Codestral-22B-v0.1"]
langs = ["py", "js", "go", "cpp", "rb"]
# langs = ["py"] #, "cpp", "py", "go", "rb"]
fc_test = False
DEBUG = False

configs = []
for model in models:
    for lang in langs:
        model_name = model.split("/")[-1]
        # fc_dir = "fc_baseline_test" if fc_test else "fc_baseline"
        fc_dir = "fc_baseline_nosuf"
        config = {
            "baseline_path": f"../../sec_data/all_results/{fc_dir}/{model_name}/temp_{temp}/{lang}",
            "temp": temp,
            "lang": lang,
            "model_dir": model,
            "fc_test": fc_test,
            "no_suffix": True,
        }
        configs.append(config)

gpus = None
launch_all_fc_infill(configs, gpus)
launch_all_fc_measure(configs)
